-----------------------------------------------------------------------------------------------
-- Client Lua Script for Navigator
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------
 
require "Window"
 
-----------------------------------------------------------------------------------------------
-- Navigator Module Definition
-----------------------------------------------------------------------------------------------
local Navigator = {} 
 
-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
-- e.g. local kiExampleVariableMax = 999
 
-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function Navigator:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 

    return o
end

function Navigator:Init()
	local bHasConfigureFunction = false
	local strConfigureButtonText = ""
	local tDependencies = {
		-- "UnitOrPackageName",
	}

    self.unitsAround = {}
    self.settings = {
	    TrackQuest = true,
    	TrackChallenge = true,
    	TrackMaster = false,
    	Rank = 4
	}

    self.rankNames = {
	    Apollo.GetString("TargetFrame_Fodder"),
	    Apollo.GetString("TargetFrame_Minion"),
	    Apollo.GetString("TargetFrame_Grunt"),
	    Apollo.GetString("TargetFrame_Challenger"),
	    Apollo.GetString("TargetFrame_Superior"),
	    Apollo.GetString("TargetFrame_Prime")
	}

    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end
 

-----------------------------------------------------------------------------------------------
-- Navigator OnLoad
-----------------------------------------------------------------------------------------------
function Navigator:OnLoad()
	Apollo.RegisterEventHandler("UnitCreated",   "OnUnitCreated",   self)
	Apollo.RegisterEventHandler("UnitDestroyed", "OnUnitDestroyed", self)

    -- load our form file
	self.xmlDoc = XmlDoc.CreateFromFile("Navigator.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
end

-----------------------------------------------------------------------------------------------
-- Navigator OnDocLoaded
-----------------------------------------------------------------------------------------------
function Navigator:OnDocLoaded()

	Apollo.RegisterEventHandler("InterfaceMenuListHasLoaded", "OnInterfaceMenuListHasLoaded", self)
	Apollo.RegisterEventHandler("OpenNavigatorConfiguration", "ToggleConfigurationWindow", self)


	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
	    self.wndMain = Apollo.LoadForm(self.xmlDoc, "NavigatorForm", nil, self)
		if self.wndMain == nil then
			Apollo.AddAddonErrorText(self, "Could not load the main window for some reason.")
			return
		end
		
	    self.wndMain:Show(false, true)

	    self.wndIcon = Apollo.LoadForm(self.xmlDoc, "ForIcons", nil, self)
		if self.wndIcon == nil then
			Apollo.AddAddonErrorText(self, "Could not load the main window for some reason.")
			return
		end
		
	    self.wndIcon:Show(false, true)

		self.xmlDoc = nil
		
		Apollo.RegisterSlashCommand("navigator", "OnNavigatorOn", self)

		self.timer      = ApolloTimer.Create(1.00, true, "Update", self)
		self.arrowTimer = ApolloTimer.Create(4.00, true, "UpdateArrow", self)
	end
end

function Navigator:InitTrackMaster()
	self.TM = Apollo.GetAddon("TrackMaster")
	if self.TM ~= nil then
		if self.settings.trackMasterLine == nil then
			self.settings.trackMasterLine = 1
		end
		self.trackMasterConfig = self.TM:AddToConfigMenu(self.TM.Type.Track, "Navigator", {
			CanFire = false,
			CanEnable = true,
			IsChecked = self.settings.TrackMaster,
			OnEnableChanged = ( function(isEnabled)
				self.settings.TrackMaster = isEnabled
				if not isEnabled and self.settings.trackMasterLine ~= nil then
					self.TM:SetTarget(nil, -1, self.settings.trackMasterLine)
				end
				if self.wndMain ~= nil then
					self.wndMain:FindChild("TrackMaster"):SetCheck(isEnabled)
				end
				self:UpdateArrow()
			end ),
			LineNo = self.settings.trackMasterLine,
			OnLineChanged = ( function(lineNo)
				if self.settings.trackMasterLine ~= nil then
					self.TM:SetTarget(nil, -1, self.settings.trackMasterLine)
				end
				self.settings.trackMasterLine = lineNo
				self:UpdateArrow()		
			end )
		})		
	end
end

-----------------------------------------------------------------------------------------------
-- Navigator Functions
-----------------------------------------------------------------------------------------------
-- Define general functions here

-- on SlashCommand "/navigator"
function Navigator:OnNavigatorOn()
	self.wndMain:Invoke() -- show the window
	self:LoadSettings()
end

function Navigator:CloseWindow()
	self.wndMain:Show(false)
end

function Navigator:OnInterfaceMenuListHasLoaded()
	Event_FireGenericEvent("InterfaceMenuList_NewAddOn", "Navigator", {"OpenNavigatorConfiguration", "", "IconSprites:Icon_Windows32_UI_CRB_InterfaceMenu_ChallengeLog"})
end

function Navigator:ToggleConfigurationWindow()
	if self.wndMain ~= nil and self.wndMain:IsVisible() then
		self:CloseWindow()
	else
		self:OnNavigatorOn()
	end
end

-- on timer
function Navigator:Update()
	-- Do your timer-related stuff here.
	local units = self:FilterUnits()
	local closest = Navigator:ClosestNode(units)

	if self.settings.TrackMaster and self.TM == nil then
		self:InitTrackMaster()
	end

	if closest == nil and self.TM ~= nil then
		self.TM:SetTarget(nil, -1, self.settings.trackMasterLine)	
	end

	if closest ~= self.closestTarget then
		self.closestTarget = closest
		self:UpdateArrow()
	end
	self.closestTarget = closest
end

function Navigator:UpdateArrow()
	if self.closestTarget == nil then
		return 
	end
	if self.settings.TrackMaster and self.TM ~= nil then
		self.TM:SetTarget(self.closestTarget, -1, self.settings.trackMasterLine)
	else
		if self.TM then
			self.TM:SetTarget(nil, -1, self.settings.trackMasterLine)
			self.TM = nil
		end
		self.closestTarget:ShowHintArrow()
	end
end

function Navigator:OnUnitCreated(unit)
	self:AddUnit(unit)
end

function Navigator:OnUnitDestroyed(unit)
	self:RemoveUnit(unit)
end

function Navigator:RankChanged( wndHandler, wndControl, fNewValue, fOldValue )
	self.wndMain:FindChild("RankName"):SetText(self.rankNames[fNewValue])
	self.settings.Rank = fNewValue
	self:Update()
	self:UpdateArrow()	
end

function Navigator:LoadSettings()
	if self.wndMain ~= nil then
		self.wndMain:FindChild("TrackQuest"):SetCheck(self.settings.TrackQuest)
		self.wndMain:FindChild("TrackMaster"):SetCheck(self.settings.TrackMaster)
		self.wndMain:FindChild("TrackChallenge"):SetCheck(self.settings.TrackChallenge)
		self.wndMain:FindChild("RankSliderBar"):SetValue(self.settings.Rank)
		self:RankChanged(nil, nil, self.settings.Rank, nil)	

	    if self.settings.wndMainLocation ~= nil then
			self.wndMain:MoveToLocation(WindowLocation.new(self.settings.wndMainLocation))
		end
	end
end

function Navigator:OnSave(saveLevel)
    if saveLevel ~= GameLib.CodeEnumAddonSaveLevel.Character then
		return
    end
    if self.wndMain ~= nil then
	    self.settings.wndMainLocation = self.wndMain:GetLocation():ToTable()
    end
    return self.settings
end

function Navigator:OnRestore(saveLevel, data)
	if saveLevel ~= GameLib.CodeEnumAddonSaveLevel.Character then 
		return 
	end	
	self.settings = self:TableMerge(self.settings, data)
end


function Navigator:OptionChanged( wndHandler, wndControl, eMouseButton )
	self.settings[wndControl:GetName()] = wndControl:IsChecked()
	self:Update()
	self:UpdateArrow()
end

-----------------------------------------------------------------------------------------------
-- NavigatorForm Functions
-----------------------------------------------------------------------------------------------

function Navigator:TableMerge(t1, t2)
    for k,v in pairs(t2) do
    	if type(v) == "table" then
    		if type(t1[k] or false) == "table" then
    			self:TableMerge(t1[k] or {}, t2[k] or {})
    		else
    			t1[k] = v
    		end
    	else
    		t1[k] = v
    	end
    end
    return t1
end

function Navigator:GetItemIndex(list, item)
	for i, v in ipairs(list) do
		if v == item then
			return i
		end
	end
	return nil
end

function Navigator:RemoveItemFromList(list, item)
	local index = self:GetItemIndex(list, item)
	if index ~= nil then
		table.remove(list, index)
		return true
	end		
	return false
end

function Navigator:DistanceToUnit(unit)
	local player = GameLib.GetPlayerUnit()
	
	if (player == nil) then 
		return 0 
	end 
 
	local posPlayer = player:GetPosition()
	local posTarget = unit:GetPosition()

	local nDeltaX = posTarget.x - posPlayer.x
	local nDeltaY = posTarget.y - posPlayer.y
	local nDeltaZ = posTarget.z - posPlayer.z
	
	return math.sqrt(math.pow(nDeltaX, 2) + math.pow(nDeltaY, 2) + math.pow(nDeltaZ, 2))
end

function Navigator:ClosestNode(units)
	local closestUnit = nil
	local min = 9999 
	for i, unit in ipairs(units) do
		local distance = self:DistanceToUnit(unit)
		if distance < min then
			closestUnit = unit
			min = distance
		end
	end
	return closestUnit
end

function Navigator:AddUnit(unit)
	if self:GetItemIndex(self.unitsAround, unit) == nil then
		table.insert(self.unitsAround, unit)
	end
end

function Navigator:RemoveUnit(unit)
	self:RemoveItemFromList(self.unitsAround, unit)
end

function Navigator:FilterUnits()
	local filteredUnits = {}
	for i, unit in ipairs(self.unitsAround) do
		if self:Filter(unit) then
			table.insert(filteredUnits, unit)
		end
	end
	return filteredUnits
end


function Navigator:Filter(unit)

	if unit:IsDead() or (unit:GetType() == "NonPlayer" and unit:GetRank() > self.settings.Rank) then
		return false
	end


	for idx, rewardInfo in pairs(unit:GetRewardInfo() or {}) do
		local rewardType = rewardInfo.strType
		if self.settings.TrackChallenge and rewardType == "Challenge" then
			for index, challenge in pairs(ChallengesLib.GetActiveChallengeList()) do
				if rewardInfo.idChallenge == challenge:GetId() and challenge:IsActivated() and not challenge:IsInCooldown() and not challenge:ShouldCollectReward() then
					return true
				end
			end
		end

		if self.settings.TrackQuest and rewardType == "Quest" then
			return true
		end
	end
	return false
end



-----------------------------------------------------------------------------------------------
-- Navigator Instance
-----------------------------------------------------------------------------------------------
local NavigatorInst = Navigator:new()
NavigatorInst:Init()
